-------------------------------------------------------------------------------
--
-- AES Decrypto Module
-- Written for 128-bit AES decryption project
-- Course: CSEE 4840 - Embedded System Design, Spring 2008
-- Authors: Shrivathsa Bhargav (sb2784)
--			Larry Chen (lc2454)
--			Abhinandan Majumdar (am2993)
--			Shiva Ramudit (syr9)
--
-- Last modified: 5-8-2008
-------------------------------------------------------------------------------

library IEEE;
use ieee.std_logic_1164.all;


--AES Decrypto Central Module
entity AES_decrypto is
port (
	ci 			           : in std_logic_vector(127 downto 0); --Input Cipher-Text/Key-Data
	clk     	   		   : in std_logic; --Clock
	en					   : in std_logic_vector(2 downto 0); --Enable lines. en(0) for demuxing Cipher-Text and Key-Data. en(1) for enabling Key Expansion. en(2) for enabling AES Decrypto
	pl 					   : out std_logic_vector(127 downto 0); --Plain Text
	eoc	      			   : out std_logic --End of Computation Signal
	
	);
end;


architecture AES_decrypto_arch of AES_decrypto is

component regis128 is
	port (
		i			: in std_logic_vector (127 downto 0);
		clk, en     : in std_logic;
		o 			: out std_logic_vector (127 downto 0)
		);
end component regis128;

component mux128_1 is
	port (
		i0,i1 : in std_logic_vector (127 downto 0);
		s     : in std_logic;
		o : out std_logic_vector (127 downto 0)
		);
end component mux128_1;

component inv_mixcolumns is
	port (
		it : in std_logic_vector (127 downto 0);
		ot : out std_logic_vector (127 downto 0)
		);
end component inv_mixcolumns;

component controller is
port (
	clk, en : in std_logic;
	count   : out std_logic_vector(3 downto 0);
	e0,e1,e2,e3,eoc   : out std_logic
	);
end component controller;

component inv_addroundkey is
	port (
		it : in std_logic_vector (127 downto 0);
		key : in std_logic_vector (127 downto 0);
		ot : out std_logic_vector (127 downto 0)
		);
end component inv_addroundkey;

component inv_shiftrow_subbytes is
	port (
		it : in std_logic_vector (127 downto 0);
		ot : out std_logic_vector (127 downto 0)
		);
end component inv_shiftrow_subbytes;

component demux1_2 is
	port (
		i : in std_logic_vector (127 downto 0);
		s     : in std_logic;
		o1, o2 : out std_logic_vector (127 downto 0)
		);
end component demux1_2;

component key_controller is
port (
	clk, en : in std_logic;
	count   : out std_logic_vector(3 downto 0);
	e0,e1,e2,eoc   : out std_logic
	);
end component key_controller;

component generate_roundkey is
	port (
		count : in std_logic_vector(3 downto 0);
		it : in std_logic_vector(127 downto 0);
		roundkey : out std_logic_vector(127 downto 0)
		);
end component generate_roundkey;

component write_controller is
port (
	clk, en : in std_logic;
	addr : out std_logic_vector(3 downto 0)
	);
end component write_controller;

component expansion_keys is
	port (
		clk, en : in std_logic;
		writeaddr : in std_logic_vector (3 downto 0);
		addr : in std_logic_vector (3 downto 0);
		it : in std_logic_vector (127 downto 0);
		key : out std_logic_vector (127 downto 0);
		writeComplete : out std_logic
		);
end component expansion_keys;

-- Signals for AES Module
signal temp1, temp2, temp3, temp4, temp5, temp6, key_itrn : std_logic_vector (127 downto 0);

signal count        : std_logic_vector (3 downto 0);
signal e0, e1, e2, e3       : std_logic;

-- Signals for Key Module
signal cnt        : std_logic_vector (3 downto 0);
signal writeaddr  : std_logic_vector (3 downto 0);
signal tempk1, tempk2, tempk3, tempk4, tempk5, tempk6 : std_logic_vector (127 downto 0);
signal key, cyphertext : std_logic_vector (127 downto 0);
signal ek0, ek1, ek2, ek3	: std_logic;

signal eodec, eokey	: std_logic;
signal endec, enkey : std_logic;

begin

-- Muxs for selecting enable and end of computation
enkey <= en(0);

endec <= en(1);
	
-- Eoc being multiplexed whether key-expansion or AES Decrypto is active
eoc <= eodec when en(1) = '1' else
       eokey when en(0) = '1' else
	'0';
	
--Controller for AES decrypto
CONTROL :  controller port map (
		  clk => clk, en => endec,
		  count => count, e0 => e0, e1 => e1, e2 => e2, e3 => e3, eoc => eodec);

--Mux1 for AES Decrypto
MUX1 :  mux128_1 port map(
		i0 => cyphertext, i1 => temp1,
		s => e0,
		o => temp2		
		);
		
--Module for Inverse Add Round Key		
INVADDRKEY :  inv_addroundkey port map(
		it => temp2,
		key => key_itrn,
		ot => temp3
		);
		
--Module for Inverse Mix Column
INVMIXCOL :  inv_mixcolumns port map(
		it => temp3,
		ot => temp4
);
		
-- Mux2 for AES Decrypto
MUX2 :  mux128_1 port map(
		i0 => temp3, i1 => temp4,
		s => e1,
		o => temp5
		);
		
-- Register1 for AES Decrypto. Stores the temporary iteration results
REG1 : regis128 port map(
		i => temp5,
		clk => clk, en => e2,
		o => temp6
		);
		
-- Module for Inverse Shiftrow Subbytes		
INVSHIFTROW_SUBBYTES:  inv_shiftrow_subbytes port map(
		it => temp6,
		ot => temp1
		);
		
--Register3 for AES Decrypto. Latches the final result
REG3 : regis128 port map(
      i => temp3,
      clk => clk, en => e3,
     o => pl
     );

-- For selecting between input and key
DEMUX : demux1_2 port map(
		i => ci, 
		s => en(2),
		o1 => key, o2 => cyphertext
);


--- Key Expansion Related Modules:
KEYCONTROL : key_controller port map (
		  clk => clk, en => enkey,
		  count => cnt, e0 => ek0, e1 => ek1, e2 => ek2, eoc => ek3
);

-- Mux to select between key and current for input into key generator
MUX3 : mux128_1 port map(
		i0 => key, i1 => tempk3,
		s => ek0,
		o => tempk1		
);

-- Round Key Generator
KEYGENERATE : generate_roundkey port map(
		count => cnt,
		it => tempk1,
		roundkey => tempk2
);

-- Mux to select between key and round key for writing to the expansion keys
MUX4 : mux128_1 port map(
		i0 => key, i1 => tempk3,
		s => ek2,
		o => tempk4
);

--Register4 for Key Expansion
REG4 : regis128 port map(
		i => tempk2,
		clk => clk, en => ek1,
		o => tempk3
);

--Register4 for Key Expansion
REG5 : regis128 port map(
		i => tempk4,
		clk => clk, en => enkey,
		o => tempk5
);

--WriteController for Key Expansion for writing the key for a particular iteration
WRITE_CONTROLLER1 : write_controller port map(
		clk => clk, en => ek3,
		addr => writeaddr
);

--Key Exansion that generates keys for every iteration
EXP_KEYS :  expansion_keys port map(
		clk => clk, 
		en => ek3,
		writeaddr => writeaddr,
		addr => count,
		it => tempk5,
		key => key_itrn,
		writeComplete => eokey 

);

end AES_decrypto_arch;
